--  FuBar_RaidLog - A logging mechanism for raids.
--  Copyright (C) 2008  J Robert Ray <jrobertray@gmail.com>
--
--  This program is free software: you can redistribute it and/or modify
--  it under the terms of the GNU General Public License as published by
--  the Free Software Foundation, either version 3 of the License, or
--  (at your option) any later version.
--
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--
--  You should have received a copy of the GNU General Public License
--  along with this program.  If not, see <http://www.gnu.org/licenses/>.

local lfbp = LibStub:GetLibrary("LibFuBarPlugin-3.0")
local tablet = LibStub:GetLibrary("Tablet-2.0")
local deformat = LibStub:GetLibrary("Deformat-2.0")

local current_raid_zone = nil
local current_raid_difficulty = nil
local BADGE_OF_JUSTICE = "Badge of Justice"
local EMBLEM_OF_HEROISM = "Emblem of Heroism"
local EMBLEM_OF_VALOR = "Emblem of Valor"
local current_roster = {}
local roster_cache = {}
local lootpatterns, unsortedloot, currentsort

local AceAddon = LibStub("AceAddon-3.0")
FuBar_RaidLog = AceAddon:NewAddon("RaidLog", "AceConsole-3.0", "AceEvent-3.0", "LibFuBarPlugin-3.0", "LibRockConfig-1.0")
FuBar_RaidLog:SetFuBarOption("tooltipType", "Tablet-2.0")
FuBar_RaidLog:SetFuBarOption("clickableTooltip", true)
FuBar_RaidLog:SetFuBarOption("configType", "LibRockConfig-1.0")

FuBar_RaidLog.options = {
    type = "group",
    name = "FuBar_RaidLog",
    desc = "A logging mechanism for raids",
    handler = FuBar_RaidLog,
    args = {
    }
}

function FuBar_RaidLog:OnInitialize()
    self.db = LibStub("AceDB-3.0"):New("FuBar_RaidLogDB", {realm = {raids = {}}})
    self.db:SetProfile("Default")

    -- port any old profile category data over to realm category
    if not self.db.realm.raids and self.db.profile.raids ~= nil then
        self.db.realm.raids = self.db.profile.raids
        self.db.profile.raids = nil
    end

    -- From SpecialEvents-Loot-1.0 >>> --

    local player = UnitName('player')

    -- Generate loot pattern handlers
    lootpatterns = { }

    -- Base loot handler function. Checks argument types, adds handler to table
    local lhandler = function(str, func)
            assert(type(str), "string", "First argument of lhandler() must be a string, instead, was ["..type(str).."]")
            assert(type(func) == "function" or not func, "Second argument of lhandler() must be a function, was ["..type(str).."]")
            table.insert(lootpatterns, { 0, 0, str, func })
    end

    -- Loot triggers
    local loot = function(who, what, num) self:SpecialEvents_ItemLooted(who, what, num or 1) end

    -- Add item patterns
    lhandler(LOOT_ITEM, loot)
    lhandler(LOOT_ITEM_MULTIPLE, loot)
    lhandler(LOOT_ITEM_PUSHED_SELF, function(what) loot(player, what) end)
    lhandler(LOOT_ITEM_PUSHED_SELF_MULTIPLE, function(what, num) loot(player, what, num) end)
    lhandler(LOOT_ITEM_SELF, function(what) loot(player, what) end)
    lhandler(LOOT_ITEM_SELF_MULTIPLE, function(what, num) loot(player, what, num) end)

    -- <<< From SpecialEvents-Loot-1.0 --

    -- Get our current roster
    self:RAID_ROSTER_UPDATE("fake")

    self:SetConfigTable(self.options)
end


function FuBar_RaidLog:OnEnable()
        self:RegisterEvent("RAID_ROSTER_UPDATE")
        self:RegisterEvent("RAID_INSTANCE_WELCOME")
        self:RegisterEvent("CHAT_MSG_SYSTEM")
        self:RegisterEvent("CHAT_MSG_LOOT")
end


function FuBar_RaidLog:CreateNewRaid(roster)
    if not roster then
        roster = self:CreateRoster()
    end
    local raid = {roster = roster, started = time(), ended = nil, where = current_raid_zone, difficulty = current_raid_difficulty, id = nil, closed = false, loot = {}}
    table.insert(self.db.realm.raids, 1, raid)
    return raid
end


function FuBar_RaidLog:PruneLastRaid()
    local last = self.db.realm.raids[1]
    if last ~= nil then
        -- if the last raid is closed, but it has no loot and
        -- was created less than 30 minutes ago, overwrite it
        if last.closed and self:RaidIsDisposable(last) then
            roster_cache[last.started] = nil
            table.remove(self.db.realm.raids, 1)
        end
    end
end


function FuBar_RaidLog:GetCurrentRaid()
    -- Return a raid table entry suitable for assigning loot
    local last = self.db.realm.raids[1]
    if last ~= nil then
        if not last.closed then
            return last
        end
        self:PruneLastRaid()
    end
    -- if GetNumPartyMembers() == 0 and GetNumRaidMembers() == 0 then return end
    if GetNumRaidMembers() == 0 then return end
    return self:CreateNewRaid()
end


function FuBar_RaidLog:SpecialEvents_ItemLooted(who, what, num)
    local name, _, qual = GetItemInfo(what)

    if name == BADGE_OF_JUSTICE or name == EMBLEM_OF_HEROISM or name == EMBLEM_OF_VALOR then return end
    if qual and qual < 3 then return end
    -- if GetNumPartyMembers() == 0 and GetNumRaidMembers() == 0 then return end
    if GetNumRaidMembers() == 0 then return end

    local raid = self:GetCurrentRaid()
    if not raid then return end
    table.insert(raid.loot, 1, {who = who, what = what, num = num, when = time()})
    self:UpdateFuBarPlugin()
end


function FuBar_RaidLog:CreateRoster()
    local roster = {}
    local thisn = 0

    for name,_ in pairs(current_roster) do
        roster[name] = true
        thisn = thisn + 1
    end

    return roster, thisn
end


function FuBar_RaidLog:RaidIsDisposable(raid)
    -- does not check if raid is closed!

    -- A raid can be thrown away if it has no loot, and if the player has
    -- never zoned into a raid zone, or the time since the raid started was
    -- 30 minutes ago or less.
    if (not raid.loot or #raid.loot == 0) and (not raid.where or raid.started >= time() - 30 * 60) then
        return true
    end
    return false
end


function FuBar_RaidLog:CloseRaid(raid)
    raid.closed = true
    raid.ended = time()
end


function FuBar_RaidLog:RosterLib_RosterUpdated()
    local roster, thisn = self:CreateRoster()

    local last = self.db.realm.raids[1]
    if last ~= nil and not last.closed then
        local lastn = 0
        local missing = false

        -- if raid count is same or going up, replace old raid entry if all
        -- the old members are in the new raid
        for oldk,_ in pairs(last.roster) do
            lastn = lastn + 1

            if not roster[oldk] then
                -- the new roster is missing a person who was in the raid
                -- previously, i.e., someone left or was replaced
                missing = true
            end
        end

        -- ignore when raid count goes down, is the raid disbanding?
        if thisn < lastn then
            self:CloseRaid(last)
            self:PruneLastRaid()
            self:UpdateFuBarPlugin()
            return
        end

        if not missing and thisn > lastn then
            -- someone joined the raid

            -- if there's no loot so far, and the raid was recently started,
            -- then it is probably still being formed
            if self:RaidIsDisposable(last) then
                -- add new person to current raid roster
                last.roster = roster
                roster_cache[last.started] = nil
                self:UpdateFuBarPlugin()
                return
            else
                -- start a new raid since there's different people now
                self:CloseRaid(last)
            end
        elseif not missing and thisn == lastn then
            -- same peoples, raid rearranging?
            return
        elseif missing then
            -- start a new raid
            self:CloseRaid(last)
        end
    end

    self:GetCurrentRaid(roster)
    self:UpdateFuBarPlugin()
end


-- From Roster-2.1 >>> --

------------------------------------------------
-- Unit iterator
------------------------------------------------

local UnitIterator
do
	local rmem, pmem, step, count
	local function SelfIterator()
		while step do
			local unit
			if step == 1 then
				-- STEP 1: player
				unit = "player"
				step = nil
			end
			if unit and UnitExists(unit) then return unit end
		end
	end

	local function SelfAndPartyIterator()
		while step do
			local unit
			if step <= 2 then
				unit = SelfIterator()
				if not step then step = 3 end
			elseif step == 3 then
				-- STEP 3: party units
				unit = string.format("party%d", count)
				count = count + 1
				step = count <= pmem and 3 or nil
			end
			if unit and UnitExists(unit) then return unit end
		end
	end

	local function RaidIterator()
		while step do
			local unit
			if step == 1 then
				-- STEP 1: raid units
				unit = string.format("raid%d", count)
				count = count + 1
				step = count <= rmem and 1 or nil
			end
			if unit and UnitExists(unit) then return unit end
		end
	end

	function UnitIterator()
		rmem = GetNumRaidMembers()
		step = 1
		if rmem == 0 then
			pmem = GetNumPartyMembers()
			if pmem == 0 then
				return SelfIterator, false
			else
				count = 1
				return SelfAndPartyIterator, false
			end
		else
			count = 1
			return RaidIterator, true
		end
	end
end

-- <<< From Roster-2.1 --

function FuBar_RaidLog:RAID_ROSTER_UPDATE(eventName)
    current_roster = {}
    local it, isInRaid = UnitIterator()
    for unitid in it do
        current_roster[UnitName(unitid)] = true
    end
    self:RosterLib_RosterUpdated()
end


function FuBar_RaidLog:CopyRoster(tbl)
    local ntbl = {}
    for key,val in pairs(tbl) do
        ntbl[key] = val
    end
    return ntbl
end


function FuBar_RaidLog:UpdateRaidLocation()
    local raid = self:GetCurrentRaid()
    if not raid then return end

    -- set the raid location if not already known
    if current_raid_zone ~= nil then
        local raid_id = nil

        local saved = GetNumSavedInstances()
        for ins = 1, saved do
            local saved_zone, id, remaining, difficulty = GetSavedInstanceInfo(ins)
            if saved_zone == current_raid_zone and difficulty == current_raid_difficulty then
                raid_id = id
                break
            end
        end

        if not raid.where then
            raid.where = current_raid_zone
            raid.difficulty = current_raid_difficulty
            raid.id = raid_id
            self:UpdateFuBarPlugin()
        elseif raid.where ~= current_raid_zone or raid.difficulty ~= current_raid_difficulty then
            -- loot from different zone?  raid must've gone to a
            -- new instance, clone the last raid
            self:CreateNewRaid(self:CopyRoster(raid.roster))
            self:UpdateFuBarPlugin()
        elseif not raid.id then
            raid.id = raid_id
            self:UpdateFuBarPlugin()
        end
    end
end


function FuBar_RaidLog:RAID_INSTANCE_WELCOME(eventName, zone, ttr)
    current_raid_zone = zone
    current_raid_difficulty = GetInstanceDifficulty()
    self:UpdateRaidLocation()
end


function FuBar_RaidLog:CHAT_MSG_SYSTEM(eventName, msg)
    if msg == INSTANCE_SAVED then
        self:UpdateRaidLocation()
    end
end


function FuBar_RaidLog:CHAT_MSG_LOOT(eventName, text)
    -- From SpecialEvents-Loot-1.0 >>> --

    -- Optimize by grouped/not and weight patterns by usage
    -- Check group mode (1 corresponds to ungrouped, 2 to grouped)
    local group = (GetNumPartyMembers() > 0 or GetNumRaidMembers() > 0) and 2 or 1
    -- If we're not currently sorting for our current mode, or it's been too long since last sort, sort ourpatterns for efficiency
    if not currentsort or currentsort ~= group or unsortedloot > 25 then
        unsortedloot = 0
        currentsort = group
        table.sort(lootpatterns, function(a, b) return a[group] < b[group] end)
    else
        unsortedloot = unsortedloot + 1
    end

    -- Match string against our patterns
    for i, v in ipairs(lootpatterns) do
        local pat, func = v[3], v[4]
        local m1, m2, m3, m4 = deformat(text, pat)
        -- Match found, add to counter and call pattern's handler
        if m1 then
            v[group] = v[group] + 1
            return func(m1, m2, m3, m4)
        end
    end

    -- <<< From SpecialEvents-Loot-1.0 --
end


function FuBar_RaidLog:DifficultyString(level)
    -- GetInstanceDifficulty() is returning 2 for both normal and heroic
    -- raids. It isn't clear yet if 2 is intended to indicate a 10-man raid.
    -- http://forums.worldofwarcraft.com/thread.html?topicId=13909293377&sid=1
    if level == 3 then
        return " (Heroic)"
    else
        return ""
    end
end


function FuBar_RaidLog:OnUpdateFuBarText()
        local last = self.db.realm.raids[1]
        if last ~= nil and not last.closed and last.where ~= nil then
            self:SetFuBarText("Raid In Progress: " .. last.where .. self:DifficultyString(last.difficulty))
        elseif last ~= nil and not last.closed and not last.where then
            self:SetFuBarText("Raid Forming")
        else
            self:SetFuBarText("No Raid")
        end
end


function FuBar_RaidLog:SortRoster(roster)
    local t = {}
    local n = 0
    for c,_ in pairs(roster) do
        table.insert(t, c)
        n = n + 1
    end
    table.sort(t)
    return t, n
end


function FuBar_RaidLog:OnUpdateFuBarTooltip()
        local cutoff = time() - 86400 * 7
        for i,raid in ipairs(self.db.realm.raids) do
            if raid.started ~= nil and raid.started < cutoff then
                break
            end

            local nextraid = self.db.realm.raids[i+1]
            if nextraid ~= nil and raid.started ~= nil and nextraid.started ~= nil and nextraid.started < raid.started - 86400 then
                nextraid = nil
            end

            -- raid header
            local cat = tablet:AddCategory("columns", 3)

            local len = 0
            if raid.started ~= nil then
                if raid.ended ~= nil then
                    len = (raid.ended - raid.started) / 60
                else
                    len = (time() - raid.started) / 60
                end
            end

            cat:AddLine(
                    "text", date("%x", raid.started) .. string.format(" (%d min)", len),
                    "text2", (raid.where or "Unknown") .. self:DifficultyString(raid.difficulty),
                    "text3", raid.id or "--"
            )

            cat = tablet:AddCategory("columns", 1)

            local cache = roster_cache[raid.started]
            if not cache then
                cache = {}

                local joined = {}
                local joined2 = {}
                local left = {}
                local left2 = {}

                if nextraid ~= nil then
                    for lastc,_ in pairs(nextraid.roster) do
                        if not raid.roster[lastc] then
                            table.insert(left, lastc)
                        end
                    end
                end

                table.sort(left)

                for c,_ in pairs(raid.roster) do
                    if nextraid ~= nil and not nextraid.roster[c] then
                        table.insert(joined, c)
                        joined2[c] = true
                    end
                end

                local sorted_roster, raidsize = self:SortRoster(raid.roster)
                if #left > 0 and #left < raidsize then
                    for _,c in ipairs(left) do
                        table.insert(sorted_roster, c)
                        left2[c] = true
                    end
                end

                local show_joined = #joined > 0 and #joined < raidsize

                local r = ""
                local acc = 0
                for _,c in ipairs(sorted_roster) do
                    local thisl = #c

                    if left2[c] then
                        c = string.format("|cffff0000-%s|r", c)
                    elseif show_joined and joined2[c] then
                        c = string.format("|cff00ff00+%s|r", c)
                    end

                    if acc > 0 and acc + thisl + 2 > 78 then
                        table.insert(cache,r)
                        r = ""
                        acc = 0
                    end

                    if r ~= "" then
                        r = r..", "..c
                        acc = acc + thisl + 2
                    else
                        r = c
                        acc = acc + thisl
                    end
                end
                table.insert(cache,r)

                roster_cache[raid.started] = cache
            end

            for _,row in ipairs(cache) do
                cat:AddLine("text", row)
            end

            -- loot list
            cat = tablet:AddCategory("columns", 2)
            if raid.loot ~= nil and #raid.loot > 0 then
                for _,loot in ipairs(raid.loot) do
                    cat:AddLine(
                        "text", loot.what,
                        "text2", loot.who,
                        "justify", "LEFT",
                        "justify2", "RIGHT",
                        "hasCheck", true,
                        "checked", true,
                        "checkIcon", GetItemIcon(loot.what),
                        "func", self.TooltipClick,
                        "arg1", self,
                        "arg2", loot.what,
                        "onEnterFunc", self.SetTooltip,
                        "onEnterArg1", self,
                        "onEnterArg2", loot.what,
                        "onLeaveFunc", GameTooltip.Hide,
                        "onLeaveArg1", GameTooltip
                    )
                end
            else
                cat:AddLine(
                    "text", "No Loot",
                    "text2", ""
                )
            end
        end
end

function FuBar_RaidLog:SetTooltip(item)
    local tipFrame = GameTooltip

    GameTooltip_SetDefaultAnchor(tipFrame, this)

    tipFrame:SetHyperlink(item)

    tipFrame:Show()
end

function FuBar_RaidLog:TooltipClick(item)
    if ChatFrameEditBox:IsVisible() then
        ChatFrameEditBox:Insert(item)
    end
end
